#!/usr/bin/env -S deno run --allow-env --allow-import --allow-net --allow-read
import Surreal from 'jsr:@surrealdb/surrealdb';
import { load } from 'jsr:@std/dotenv';
import { join } from 'jsr:@std/path';
import { parseArgs } from 'jsr:@std/cli/parse-args';

interface DatabaseConfig {
	DB_URL: string;
	NAMESPACE: string;
	DATABASE: string;
	USERNAME: string;
	PASSWORD: string;
}

interface Command {
	name: string;
	description: string;
	cliCommand?: string;
	execute: (db: Surreal, args?: Record<string, unknown>) => Promise<void>;
	requiresInput?: boolean;
	inputPrompt?: string;
}

const env = await load();

const config: DatabaseConfig = {
	DB_URL: env.DB_URL || 'http://localhost:8000/rpc',
	NAMESPACE: env.DB_NS || 'grady',
	DATABASE: env.DB_NAME || 'dev',
	USERNAME: env.DB_USER || 'root',
	PASSWORD: env.DB_PASS || 'root'
};

async function connectToDatabase() {
	const db = new Surreal();
	await db.connect(config.DB_URL);
	await db.signin({
		username: config.USERNAME,
		password: config.PASSWORD
	});
	await db.use({ namespace: config.NAMESPACE, database: config.DATABASE });
	return db;
}

const commands: Command[] = [
	{
		name: 'Check Status',
		description: 'Check if the database is reachable',
		cliCommand: 'status',
		async execute(db: Surreal) {
			try {
				const timeout = new Promise((_, reject) =>
					setTimeout(() => reject(new Error('Timeout')), 5000)
				);
				await Promise.race([db.ping(), timeout]);

				process.exit(0);
			} catch (error) {
				console.error('Database not reachable', error);
				process.exit(1);
			}
		}
	},
	{
		name: 'Init',
		description: 'Initialize database for development environment',
		cliCommand: 'init',
		async execute(db: Surreal) {
			try {
				// Check if we're already initialized
				const users = await db.select('user');
				// console.log(users);
				if (users !== undefined && users.length > 0) {
					console.warn('Already initialized');
					return;
				}

				// Init schema
				const schemaDir = join(Deno.cwd(), 'db/migrations');
				for await (const dirEntry of Deno.readDir(schemaDir)) {
					if (dirEntry.name.endsWith('.surql')) {
						console.log(`Applying migration: ${dirEntry.name}`);
						const filePath = join(schemaDir, dirEntry.name);
						const sql = await Deno.readTextFile(filePath);
						await db.query(sql);
						console.log(`Migration applied: ${dirEntry.name}`);
					}
				}
				console.log('All migrations applied successfully');

				// Init three default users
				await db.query('DEFINE USER dev ON ROOT PASSWORD "dev" ROLES OWNER;');
				await db.signup({
					access: 'user',
					variables: {
						name: 'tutor',
						email: 'tutor@grady.org',
						roles: ['tutor'],
						password: 'tutor'
					}
				});
				await db.signup({
					access: 'user',
					variables: {
						name: 'reviewer',
						email: 'reviewer@grady.org',
						roles: ['reviewer'],
						password: 'reviewer'
					}
				});
				console.info('Initialized dev DB');
			} catch (error) {
				console.error('Failed to initialize DB: ', error);
			}
		}
	},
	{
		name: 'Delete All Tables',
		description: 'Remove all tables from the current database',
		cliCommand: 'delete-all',
		async execute(db: Surreal) {
			try {
				const result: any[] = await db.query('INFO FOR DB');
				const tables = Object.keys(result[0].tables);
				for (const table of tables) {
					await db.query(`REMOVE TABLE ${table}`);
					console.log(`Deleted table: ${table}`);
				}
				console.log('All tables deleted successfully');
			} catch (error) {
				console.error('Error deleting all tables:', error);
			}
		}
	},
	{
		name: 'Delete Specific Table',
		description: 'Remove a specific table from the database',
		cliCommand: 'delete-table',
		requiresInput: true,
		async execute(db: Surreal) {
			try {
				const tables = await listTables(db);
				if (tables.length === 0) {
					console.log('No tables found in the database.');
					return;
				}

				const selectedTable = await promptForTableSelection(tables);
				if (selectedTable) {
					await db.query(`REMOVE TABLE ${selectedTable}`);
					console.log(`Deleted table: ${selectedTable}`);
				} else {
					console.log('Invalid table selection.');
				}
			} catch (error) {
				console.error('Error deleting table:', error);
			}
		}
	},
	{
		name: 'Show Database Info',
		description: 'Display information about the current database',
		cliCommand: 'info',
		async execute(db: Surreal) {
			try {
				console.log('Database Information:');
				console.log('---------------------');
				const dbInfo = await db.query('INFO FOR DB');
				console.log(JSON.stringify(dbInfo, null, 2));
			} catch (error) {
				console.error('Error retrieving database info:', error);
			}
		}
	},
	{
		name: 'Drop Database',
		description: 'Remove the current database',
		cliCommand: 'drop',
		async execute(db: Surreal) {
			try {
				await db.query(`REMOVE DATABASE ${config.DATABASE}`);
				console.log('Database dropped successfully');
			} catch (error) {
				console.error('Error dropping database:', error);
			}
		}
	},
	{
		name: 'Execute File',
		description: 'Execute a SurrealQL file',
		cliCommand: 'execute',
		requiresInput: true,
		inputPrompt: 'Enter the path to the SurrealQL file:',
		async execute(db: Surreal, args?: Record<string, unknown>) {
			try {
				let filePath: string;

				if (args?.file) {
					filePath = args.file as string;
				} else if (args?.seed) {
					// Maintain backward compatibility with 'seed' command
					filePath = args.seed as string;
				} else {
					const buf = new Uint8Array(1024);
					console.log(this.inputPrompt);
					const n = await Deno.stdin.read(buf);
					if (n === null) throw new Error('No input provided');
					filePath = new TextDecoder().decode(buf.subarray(0, n)).trim();
				}

				if (!filePath) {
					throw new Error('No file path provided');
				}

				if (!filePath.startsWith('/')) {
					filePath = join(Deno.cwd(), filePath);
				}

				try {
					await Deno.stat(filePath);
				} catch {
					throw new Error(`File not found: ${filePath}`);
				}

				console.log(`Executing file: ${filePath}`);
				const sql = await Deno.readTextFile(filePath);

				try {
					const result = await db.query(sql);
					console.log('Result:', JSON.stringify(result, null, 2));
				} catch (error) {
					console.error(`Error executing sql:`, error);
					if (args?.continueOnError !== true) {
						throw new Error('Execution stopped due to error');
					}
				}

				console.log('\nFile execution completed successfully');
			} catch (error) {
				console.error('Error executing file:', error);
				throw error;
			}
		}
	},
	{
		name: 'Execute Query',
		description: 'Execute a custom SQL query',
		cliCommand: 'query',
		requiresInput: true,
		inputPrompt: 'Enter your query (press Ctrl+D or Ctrl+Z when done):',
		async execute(db: Surreal, args?: Record<string, unknown>) {
			try {
				let sql: string;
				if (args?.query) {
					sql = args.query as string;
				} else {
					sql = await readMultilineInput();
				}

				if (sql.trim()) {
					console.log('\nExecuting query:');
					console.log(sql);
					console.log('\nResult:');
					const result = await db.query(sql);
					console.log(JSON.stringify(result, null, 2));
				}
			} catch (error) {
				console.error('Error executing query:', error);
			}
		}
	}
];

async function listTables(db: Surreal): Promise<string[]> {
	try {
		const result = (await db.query('INFO FOR DB')) as any;
		return Object.keys(result[0].tables);
	} catch (error) {
		console.error('Error listing tables:', error);
		return [];
	}
}

async function readMultilineInput(): Promise<string> {
	const chunks: string[] = [];
	const decoder = new TextDecoder();
	const buf = new Uint8Array(1024);

	while (true) {
		const n = await Deno.stdin.read(buf);
		if (n === null) break;
		chunks.push(decoder.decode(buf.subarray(0, n)));
	}
	return chunks.join('');
}

async function promptForTableSelection(tables: string[]): Promise<string | null> {
	console.log('Available tables:');
	tables.forEach((table, index) => {
		console.log(`${index + 1}. ${table}`);
	});
	console.log('Enter the number of the table you want to delete: ');

	const buf = new Uint8Array(1024);
	const n = await Deno.stdin.read(buf);
	if (n === null) return null;

	const input = new TextDecoder().decode(buf.subarray(0, n)).trim();
	const index = parseInt(input) - 1;
	return index >= 0 && index < tables.length ? tables[index] : null;
}

async function promptForCommand(): Promise<Command | null> {
	console.log('\nAvailable commands:');
	commands.forEach((cmd, index) => {
		console.log(`${index + 1}. ${cmd.name} - ${cmd.description}`);
	});
	console.log(`${commands.length + 1}. Exit`);
	console.log('Enter the number of your choice: ');

	const buf = new Uint8Array(1024);
	const n = await Deno.stdin.read(buf);
	if (n === null) return null;

	const input = parseInt(new TextDecoder().decode(buf.subarray(0, n)).trim());
	if (input === commands.length + 1) {
		return null;
	}
	return input > 0 && input <= commands.length ? commands[input - 1] : null;
}

async function main() {
	const args = parseArgs(Deno.args, {
		string: ['namespace', 'database', 'file', 'query'],
		boolean: ['continueOnError']
	});

	if (args.namespace) config.NAMESPACE = args.namespace;
	if (args.database) config.DATABASE = args.database;

	console.log(`Using namespace: ${config.NAMESPACE}, database: ${config.DATABASE}`);

	const cliCommand = args._[0]?.toString();

	if (cliCommand) {
		// cli mode
		const command = commands.find((cmd) => cmd.cliCommand === cliCommand);
		if (command) {
			const db = await connectToDatabase();
			try {
				await command.execute(db, args);
			} finally {
				await db.close();
			}
		} else {
			console.log('Available commands:');
			commands.forEach((cmd) => {
				if (cmd.cliCommand) {
					console.log(`  ${cmd.cliCommand} - ${cmd.description}`);
				}
			});
			console.log(
				'\nOptional flags: --namespace <namespace>, --database <database>, --seed <path>, --query <sql>'
			);
		}
	} else {
		// interactive mode
		while (true) {
			const selectedCommand = await promptForCommand();
			if (!selectedCommand) {
				console.log('Exiting...');
				break;
			}

			const db = await connectToDatabase();
			try {
				await selectedCommand.execute(db);
			} finally {
				await db.close();
			}
			console.log('\n');
		}
	}
}

if (import.meta.main) {
	main().catch((error) => {
		console.error(error);
		Deno.exit(1);
	});
}
