#!/usr/bin/env bun
import Surreal from 'surrealdb';
import { config as loadEnv } from 'dotenv';
import { join } from 'path';
import minimist from 'minimist';
import { promises as fs } from 'fs';
import readline from 'readline';

loadEnv();

interface DatabaseConfig {
	DB_URL: string;
	NAMESPACE: string;
	DATABASE: string;
	USERNAME: string;
	PASSWORD: string;
}

interface Command {
	name: string;
	description: string;
	cliCommand?: string;
	execute: (db: Surreal, args?: Record<string, unknown>) => Promise<void>;
	requiresInput?: boolean;
	inputPrompt?: string;
}

const config: DatabaseConfig = {
	DB_URL: process.env.DB_URL || 'ws://localhost:8000/rpc',
	NAMESPACE: process.env.DB_NS || 'grady',
	DATABASE: process.env.DB_NAME || 'dev',
	USERNAME: process.env.DB_USER || 'root',
	PASSWORD: process.env.DB_PASS || 'root'
};

const rl = readline.createInterface({
	input: process.stdin,
	output: process.stdout
});

async function connectToDatabase() {
	const db = new Surreal();
	await db.connect(config.DB_URL);
	await db.signin({
		username: config.USERNAME,
		password: config.PASSWORD
	});
	await db.use({ namespace: config.NAMESPACE, database: config.DATABASE });
	return db;
}

const commands: Command[] = [
	{
		name: 'Init',
		description: 'Initialize database for development environment',
		cliCommand: 'init',
		async execute(db: Surreal) {
			try {
				// Check if we're already initialized
				type DatabaseSchema = {
					accesses: Record<string, never>;
					analyzers: Record<string, never>;
					configs: Record<string, never>;
					functions: Record<string, never>;
					models: Record<string, never>;
					params: Record<string, never>;
					tables: Record<string, never>;
					users: Record<string, never>;
				};

				const dbSchemaQuery = await db.query<[DatabaseSchema]>('INFO FOR DB;');
				const dbSchema = dbSchemaQuery[0];

				if (Object.keys(dbSchema.accesses).length !== 0) {
					console.warn('A schema has already been applied');
				} else {
					const schemaFile = join(process.cwd(), 'db/schema/schema.surql');
					const sql = await fs.readFile(schemaFile, 'utf8');
					await db.query(sql);
					console.log('Applied schema successfully');
				}

				const users = await db.select('user');
				if (users !== undefined && users.length > 0) {
					console.warn('Users have already been initialized');
					return;
				}

				// Init default users for development
				const result = await db.query(`
					CREATE user SET 
						name = 'reviewer',
						email = 'reviewer@grady.org',
						password_hash = crypto::argon2::generate('reviewer'),
						enabled = true,
						roles = ['reviewer'];

					CREATE user SET 
						name = 'tutor',
						email = 'tutor@grady.org',
						password_hash = crypto::argon2::generate('tutor'),
						enabled = true,
						roles = ['tutor'];
				`);
				console.info('Initialized dev DB');
			} catch (error) {
				console.error('Failed to initialize DB: ', error);
			}
		}
	},
	{
		name: 'Check Status',
		description: 'Check if the database is reachable',
		cliCommand: 'status',
		async execute(db: Surreal) {
			try {
				const timeout = new Promise((_, reject) =>
					setTimeout(() => reject(new Error('Timeout')), 5000)
				);
				await Promise.race([db.ping(), timeout]);
				process.exit(0);
			} catch (error) {
				console.error('Database not reachable', error);
				process.exit(1);
			}
		}
	},
	{
		name: 'Apply Migrations',
		description: 'Apply all migration files from the db/migrations directory',
		cliCommand: 'migrate',
		async execute(db: Surreal) {
			try {
				const schemaDir = join(process.cwd(), 'db/migrations');
				const dir = await fs.opendir(schemaDir);
				for await (const dirEntry of dir) {
					if (dirEntry.name.endsWith('.surql')) {
						console.log(`Applying migration: ${dirEntry.name}`);
						const filePath = join(schemaDir, dirEntry.name);
						const sql = await fs.readFile(filePath, 'utf8');
						await db.query(sql);
						console.log(`Migration applied: ${dirEntry.name}`);
					}
				}
				console.log('All migrations applied successfully');
			} catch (error) {
				console.error('Error applying migrations:', error);
			}
		}
	},
	{
		name: 'Delete All Tables',
		description: 'Remove all tables from the current database',
		cliCommand: 'delete-all',
		async execute(db: Surreal) {
			try {
				const result: any[] = await db.query('INFO FOR DB');
				const tables = Object.keys(result[0].tables);
				for (const table of tables) {
					await db.query(`REMOVE TABLE ${table}`);
					console.log(`Deleted table: ${table}`);
				}
				console.log('All tables deleted successfully');
			} catch (error) {
				console.error('Error deleting all tables:', error);
			}
		}
	},
	{
		name: 'Delete Specific Table',
		description: 'Remove a specific table from the database',
		cliCommand: 'delete-table',
		requiresInput: true,
		async execute(db: Surreal) {
			try {
				const tables = await listTables(db);
				if (tables.length === 0) {
					console.log('No tables found in the database.');
					return;
				}

				const selectedTable = await promptForTableSelection(tables);
				if (selectedTable) {
					await db.query(`REMOVE TABLE ${selectedTable}`);
					console.log(`Deleted table: ${selectedTable}`);
				} else {
					console.log('Invalid table selection.');
				}
			} catch (error) {
				console.error('Error deleting table:', error);
			}
		}
	},
	{
		name: 'Show Database Info',
		description: 'Display information about the current database',
		cliCommand: 'info',
		async execute(db: Surreal) {
			try {
				console.log('Database Information:');
				console.log('---------------------');
				const dbInfo = await db.query('INFO FOR DB');
				console.log(JSON.stringify(dbInfo, null, 2));
			} catch (error) {
				console.error('Error retrieving database info:', error);
			}
		}
	},
	{
		name: 'Drop Database',
		description: 'Remove the current database',
		cliCommand: 'drop',
		async execute(db: Surreal) {
			try {
				await db.query(`REMOVE DATABASE ${config.DATABASE}`);
				console.log('Database dropped successfully');
			} catch (error) {
				console.error('Error dropping database:', error);
			}
		}
	},
	{
		name: 'Execute File',
		description: 'Execute a SurrealQL file',
		cliCommand: 'execute',
		requiresInput: true,
		inputPrompt: 'Enter the path to the SurrealQL file:',
		async execute(db: Surreal, args?: Record<string, unknown>) {
			try {
				let filePath: string;

				if (args?.file) {
					filePath = args.file as string;
				} else if (args?.seed) {
					filePath = args.seed as string;
				} else {
					filePath = await new Promise((resolve) => {
						rl.question(this.inputPrompt + '\n', resolve);
					});
				}

				if (!filePath) throw new Error('No file path provided');
				if (!filePath.startsWith('/')) {
					filePath = join(process.cwd(), filePath);
				}

				try {
					await fs.access(filePath);
				} catch {
					throw new Error(`File not found: ${filePath}`);
				}

				console.log(`Executing file: ${filePath}`);
				const sql = await fs.readFile(filePath, 'utf8');

				try {
					const result = await db.query(sql);
					console.log('Result:', JSON.stringify(result, null, 2));
				} catch (error) {
					console.error(`Error executing sql:`, error);
					if (args?.continueOnError !== true) {
						throw new Error('Execution stopped due to error');
					}
				}

				console.log('\nFile execution completed successfully');
			} catch (error) {
				console.error('Error executing file:', error);
				throw error;
			}
		}
	},
	{
		name: 'Execute Query',
		description: 'Execute a custom SQL query',
		cliCommand: 'query',
		requiresInput: true,
		inputPrompt: 'Enter your query (press Ctrl+D or Ctrl+Z when done):',
		async execute(db: Surreal, args?: Record<string, unknown>) {
			try {
				let sql: string;
				if (args?.query) {
					sql = args.query as string;
				} else {
					console.log(this.inputPrompt);
					sql = await readMultilineInput();
				}

				if (sql.trim()) {
					console.log('\nExecuting query:');
					console.log(sql);
					console.log('\nResult:');
					const result = await db.query(sql);
					console.log(JSON.stringify(result, null, 2));
				}
			} catch (error) {
				console.error('Error executing query:', error);
			}
		}
	}
];

async function listTables(db: Surreal): Promise<string[]> {
	try {
		const result = (await db.query('INFO FOR DB')) as any;
		return Object.keys(result[0].tables);
	} catch (error) {
		console.error('Error listing tables:', error);
		return [];
	}
}

async function readMultilineInput(): Promise<string> {
	const chunks: Buffer[] = [];
	for await (const chunk of process.stdin) {
		chunks.push(chunk);
	}
	return Buffer.concat(chunks).toString().trim();
}

async function promptForTableSelection(tables: string[]): Promise<string | null> {
	console.log('Available tables:');
	tables.forEach((table, index) => {
		console.log(`${index + 1}. ${table}`);
	});

	const input = await new Promise<string>((resolve) => {
		rl.question('Enter the number of the table you want to delete: ', resolve);
	});

	const index = parseInt(input.trim()) - 1;
	return index >= 0 && index < tables.length ? tables[index] : null;
}

async function promptForCommand(): Promise<Command | null> {
	console.log('\nAvailable commands:');
	commands.forEach((cmd, index) => {
		console.log(`${index + 1}. ${cmd.name} - ${cmd.description}`);
	});
	console.log(`${commands.length + 1}. Exit`);

	const input = await new Promise<string>((resolve) => {
		rl.question('Enter the number of your choice: ', resolve);
	});

	const choice = parseInt(input.trim());
	if (choice === commands.length + 1) return null;
	return choice > 0 && choice <= commands.length ? commands[choice - 1] : null;
}

async function main() {
	const args = minimist(process.argv.slice(2), {
		string: ['namespace', 'database', 'file', 'query'],
		boolean: ['continueOnError']
	});

	if (args.namespace) config.NAMESPACE = args.namespace;
	if (args.database) config.DATABASE = args.database;

	console.log(`Using namespace: ${config.NAMESPACE}, database: ${config.DATABASE}`);

	const cliCommand = args._[0]?.toString();

	if (cliCommand) {
		const command = commands.find((cmd) => cmd.cliCommand === cliCommand);
		if (command) {
			const db = await connectToDatabase();
			try {
				await command.execute(db, args);
			} finally {
				await db.close();
				rl.close();
			}
		} else {
			console.log('Available commands:');
			commands.forEach((cmd) => {
				if (cmd.cliCommand) console.log(`  ${cmd.cliCommand} - ${cmd.description}`);
			});
			console.log(
				'\nOptional flags: --namespace <namespace>, --database <database>, --seed <path>, --query <sql>'
			);
		}
	} else {
		while (true) {
			const selectedCommand = await promptForCommand();
			if (!selectedCommand) {
				console.log('Exiting...');
				rl.close();
				break;
			}

			const db = await connectToDatabase();
			try {
				await selectedCommand.execute(db);
			} finally {
				await db.close();
			}
			console.log('\n');
		}
	}
}

if (import.meta.main) {
	main().catch((error) => {
		console.error(error);
		process.exit(1);
	});
}
