{ pkgs, lib, config, inputs, ... }:

let
  pkgs-unstable = import inputs.nixpkgs-unstable { system = pkgs.stdenv.system; };
  dev-deps = with pkgs-unstable; [
    git
    vscode-langservers-extracted
    nodePackages.typescript-language-server
    svelte-language-server
    tailwindcss-language-server
    prettierd
  ];

  # dev db config
  surreal = {
    port = toString 8000;
    host = "localhost";
    user = "root";
    pass = "root";
    dataDir = "run/data/surreal";
    dbFile = "dev.db";
  };
in
{
  tasks."grady:build" = { exec = "${config.languages.javascript.bun.package}/bin/bun --bun run build"; }; # before = [ "devenv:enterShell" ]; };

  dotenv.enable = true;
  # dotenv.filename = [ ".env" ];

  cachix.enable = true;
  # cachix.pull = [ "grady" ];
  cachix.push = "grady"; # push/pull. TODO only push on CI

  difftastic.enable = true;

  # containers.processes.registry = "docker://docker.gitlab.gwdg.de/grady-corp/grady/";
  # containers.processes.defaultCopyArgs = [
  #   "--dest-creds"
  #   "USERTODO:TOKENTODO"
  # ];

  # maybe like this?
  containers.prod.name = "grady-prod";
  containers.prod.copyToRoot = ./build;
  containers.prod.startupCommand = "bun --bun index.js";

  # https://devenv.sh/basics/
  env = {
    # SURREAL_NS = "grady";
    # SURREAL_DB = "dev";
    # SURREAL_USER = surreal.user;
    # SURREAL_PASS = surreal.pass;
    # SURREAL_HOST = surreal.host;
    # SURREAL_PORT = surreal.port;
    # SURREAL_DATA_DIR = surreal.dataDir;
    # SURREAL_DATA_DB = surreal.dbFile;
  };

  # https://devenv.sh/packages/
  packages = [ pkgs.openssl inputs.surrealdb-git.packages.${pkgs.stdenv.system}.default ] ++ lib.optionals (!config.container.isBuilding) dev-deps;

  # https://devenv.sh/languages/
  # languages.rust.enable = true;
  languages = {
    javascript = {
      enable = true;
      # npm.enable = true;
      # pnpm.enable = true;
      bun.enable = true;
      bun.install.enable = false;
    };
    typescript.enable = true;
    deno.enable = false;
  };

  # https://devenv.sh/processes/
  process.manager.implementation = "process-compose";
  processes = {
    surrealdb = {
      exec = ''
        surreal start \
        --bind 127.0.0.1:8000 \
        --user ${surreal.user} \
        --pass ${surreal.pass} \
        --log debug \
        --import-file db/schema/schema.surql \
        rocksdb://${surreal.dataDir}/${surreal.dbFile}
      '';
      process-compose = {
        is_daemon = true;
        readiness_probe = {
          # exec = {
          #   command = "curl -I http://localhost:8000/status";
          # };
          http_get = {
            scheme = "http";
            host = "127.0.0.1";
            port = 8000;
            path = "/status";
          };
          initial_delay_seconds = 5;
          period_seconds = 5;
          timeout_seconds = 5;
          success_threshold = 1;
          failure_threshold = 10;
        };
        liveness_probe = {
          # exec = {
          #   command = "curl -I http://localhost:8000/health";
          # };
          http_get = {
            scheme = "http";
            host = "127.0.0.1";
            port = 8000;
            path = "/health";
          };
          initial_delay_seconds = 5;
          period_seconds = 5;
          timeout_seconds = 5;
          success_threshold = 1;
          failure_threshold = 5;
        };
      };
    };
    init = {
      exec = "./db/manage.ts init";
      # exec = "echo noop";
      process-compose = {
        depends_on.surrealdb.condition = "process_healthy";
      };
    };
    grady = {
      exec = ''
        bun install
        bun --bun run dev

        # TODO prod
        # bun --bun run build
        # bun build/index.js
      '';
      process-compose = {
        depends_on.init.condition = "process_completed_successfully";
      };
    };
  };

  # https://devenv.sh/services/
  # services.postgres.enable = true;

  # https://devenv.sh/scripts/
  # enterShell = ''
  #   git --version
  # '';

  # https://devenv.sh/tasks/
  tasks = {
    "app:dev" = {
      exec = "pnpm dev";
      # status = "./db/manage.ts status";
    };
    "db:migrate" = {
      exec = "./db/manage.ts migrate";
      status = "./db/manage.ts status";
    };
    "db:info" = {
      exec = "./db/manage.ts info";
      status = "./db/manage.ts status";
    };
    "db:query" = {
      exec = ''
        if [ -z "$QUERY" ]; then
          ./db/manage.ts query
        else
          ./db/manage.ts query --query "$QUERY"
        fi
      '';
      status = "./db/manage.ts status";
    };
  };

  # https://devenv.sh/tests/
  enterTest = ''
    echo "Running tests"
    git --version | grep --color=auto "${pkgs.git.version}"
  '';

  # https://devenv.sh/pre-commit-hooks/
  # pre-commit.hooks.shellcheck.enable = true;
  pre-commit.hooks = {
    eslint.enable = true;
    prettier.enable = true;
    prettier.settings = {
      configPath = ".prettierrc";
    };
  };

  # See full reference at https://devenv.sh/reference/options/
}
