import { z } from 'zod';
import { getDb } from '$lib/surreal.svelte';
import {
	moduleCreateSchema,
	assignmentCreateSchema,
	userCreateSchema,
	submissionCreateSchema,
	studentCreateSchema,
	type ModuleCreate,
	type AssignmentCreate,
	type StudentCreate,
	type SubmissionCreate
} from '$lib/surreal/schema';

export const parseAndCreateExamData = async (data: unknown) => {
	const validatedImport = importSchema.parse(data);
	const validatedData = validatedImport.data;
	// const validatedData = examDataSchema.parse(data);

	const db = await getDb();
	// TODO use proper authentication, restrict to admin users
	// TEMP
	await db.signin({ username: 'root', password: 'root' });

	let module = moduleCreateSchema.parse({
		...validatedData.module,
		reference: validatedData.module.module_reference,
		ungraded: validatedData.module.pass_only
	});

	const [moduleRecord] = await db.create<ModuleCreate>('module', module);

	const moduleId = moduleRecord?.id;
	console.log(moduleId);

	const assignments = validatedData.submission_types.map((st) =>
		assignmentCreateSchema.parse({
			...st,
			module: moduleId
		})
	);

	const assignmentRecords = await db.insert<AssignmentCreate>('assignment', assignments);

	const assignmentMap = new Map(assignmentRecords.map((a) => [a.name, a.id]));

	// TODO ignoring exercise_groups on import for now, since I've seen no data set where that field is actually being used
	const studentRecords = await Promise.all(
		validatedData.students.map(async (s) => {
			const studentData = studentCreateSchema.parse({ alias: s.identifier });
			const [studentRecord] = await db.create<StudentCreate>('student', studentData);
			return {
				studentId: studentRecord.id,
				originalData: s
			};
		})
	);

	const submissions = studentRecords.flatMap(({ studentId, originalData }) =>
		originalData.submissions.map((sub) =>
			submissionCreateSchema.parse({
				...sub,
				source_code: sub.source_code ?? undefined, // from null to option/NONE
				in: studentId,
				out: assignmentMap.get(sub.type)
			})
		)
	);

	const submissionsResult = await db.insert_relation<SubmissionCreate>('submission', submissions);

	return {
		moduleId,
		assignmentIds: Array.from(assignmentMap.values()),
		studentIds: studentRecords.map((r) => r.studentId),
		submissionIds: submissionsResult.map((s) => s.id)
	};
};

const examDataSchema = z
	.object({
		module: z
			.object({
				module_reference: z.string(),
				total_score: z.number(),
				pass_score: z.number(),
				pass_only: z.boolean()
			})
			.passthrough(),
		submission_types: z.array(
			z
				.object({
					name: z.string(),
					full_score: z.number(),
					description: z.string(),
					solution: z.string(),
					programming_language: z.string()
				})
				.passthrough()
		),
		students: z.array(
			z
				.object({
					fullname: z.string(),
					identifier: z.string(),
					username: z.string(),
					email: z.string(),
					exercise_groups: z.array(z.string()),
					submissions: z.array(
						z
							.object({
								code: z.string(),
								source_code: z.string().nullable(),
								type: z.string(),
								tests: z.array(
									z
										.object({
											name: z.string(),
											annotation: z.string(),
											label: z.string()
										})
										.passthrough()
								)
							})
							.passthrough()
					)
				})
				.passthrough()
		)
	})
	.passthrough();

const importSchema = z.object({
	meta: z.any(),
	data: examDataSchema
});
