import type { Roles, User } from '$types';
import type { RoleRequirement, RouteConfig, GuardResult, RouteGuard } from './types';
import { isRedirect, redirect, type Handle } from '@sveltejs/kit';
import { DEFAULT_CONFIG } from './config';

export class RouteMatchError extends Error {
	constructor(message: string) {
		super(message);
		this.name = 'RouteMatchError';
	}
}

interface RouteMatch {
	route: RouteGuard;
	specificity: number;
	params?: Record<string, string>;
}

function getRouteSpecificity(pattern: string | RegExp): number {
	if (pattern instanceof RegExp) {
		return 100;
	}

	const segments = pattern.split('/').filter(Boolean);
	return segments.length * 10 + (pattern.endsWith('/') ? 0 : 1);
}

function matchRoute(pathname: string, pattern: string | RegExp): boolean {
	try {
		if (pattern instanceof RegExp) {
			return pattern.test(pathname);
		}

		const normalizedPath = pathname.endsWith('/') ? pathname : `${pathname}/`;
		const normalizedPattern = pattern.endsWith('/') ? pattern : `${pattern}/`;

		return normalizedPath.startsWith(normalizedPattern);
	} catch (error) {
		throw new RouteMatchError(`Invalid route pattern: ${pattern}`);
	}
}

function findMostSpecificRoute(pathname: string, routes: RouteGuard[]): RouteGuard | null {
	const matches: RouteMatch[] = [];

	for (const route of routes) {
		if (matchRoute(pathname, route.pattern)) {
			matches.push({
				route,
				specificity: getRouteSpecificity(route.pattern)
			});
		}
	}

	matches.sort((a, b) => b.specificity - a.specificity);
	// console.log('Matches:');
	// console.log(matches);

	return matches.length > 0 ? matches[0].route : null;
}

function checkRoleRequirement(userRoles: Roles, requirement: RoleRequirement): boolean {
	if (!userRoles) {
		return false;
	}

	if (typeof requirement === 'string') {
		return userRoles.has(requirement);
	}

	if (!requirement.roles) {
		return false;
	}

	return requirement.requireAll
		? requirement.roles.isSubsetOf(userRoles)
		: requirement.roles.isSupersetOf(userRoles);
}

function evaluateAccess(
	pathname: string,
	user: User | null,
	config: RouteConfig = DEFAULT_CONFIG
): GuardResult {
	const matchedAuthRoute = config.authRoutes.some((route) => matchRoute(pathname, route));
	if (matchedAuthRoute) {
		if (user) {
			return {
				allowed: false,
				redirectTo: redirect(303, config.defaultAuthRedirect)
			};
		}
		return { allowed: true };
	}

	const matchedRoute = findMostSpecificRoute(pathname, config.protectedRoutes);

	// console.log('Matched route:');
	// console.log(matchedRoute);

	if (!matchedRoute) {
		return { allowed: true };
	}

	if (!user) {
		return {
			allowed: false,
			redirectTo: redirect(303, `/signin?returnTo=${encodeURIComponent(pathname)}`)
		};
	}

	if (!matchedRoute.roles) {
		return { allowed: true };
	}

	const hasRequiredRoles = checkRoleRequirement(new Set(user.roles), matchedRoute.roles);

	if (!hasRequiredRoles) {
		return {
			allowed: false,
			redirectTo: redirect(
				303,
				`${config.unauthorizedRedirect}?from=${encodeURIComponent(pathname)}`
			)
		};
	}

	return { allowed: true };
}

const createGuardRoutes = (config: Partial<RouteConfig> = {}): Handle => {
	const mergedConfig = { ...DEFAULT_CONFIG, ...config };

	return async ({ event, resolve }) => {
		try {
			const pathname = event.url.pathname;
			const user = event.locals.user;

			// console.log(`Evaluating route guard for path: ${pathname}`);
			// if (user) {
			// 	console.log(`User roles: ${Array.from(user.roles).join(', ')}`);
			// }

			const { allowed, redirectTo } = evaluateAccess(pathname, user, mergedConfig);

			// console.log(`Route guard result: allowed=${allowed}, redirect=${redirectTo}`);

			if (!allowed && redirectTo) {
				throw redirectTo;
			}

			return resolve(event);
		} catch (error) {
			if (isRedirect(error)) {
				throw error;
			}
			console.error('Route guard error:', error);
			throw redirect(303, mergedConfig.unauthorizedRedirect);
		}
	};
};

export const guardRoutes = createGuardRoutes();
