import { jsonify, Surreal } from 'surrealdb';
import type { SessionInfo } from '$lib/types';
import { sessionInfoSchema } from './zod';
// import {
//   PUBLIC_DB_URL,
//   PUBLIC_DB_DB,
//   PUBLIC_DB_NS,
// } from '$env/static/public';
import { AuthCookie } from './cookies';
import { userSchema, type User } from './surreal/schema';

type DbConfig = {
	url: string;
	namespace: string;
	database: string;
};

const DEFAULT_CONFIG: DbConfig = {
	url: 'http://localhost:8000',
	namespace: 'grady',
	database: 'dev'
};

let connectionPromise: Promise<Surreal> | null = null;
let currentUser: User | null = $state(null);
let isAuthenticated = $state(false);

// export const user: User | null = $derived(currentUser)
// export const authenticated = $derived(isAuthenticated)

const connectToDatabase = async (config: DbConfig): Promise<Surreal> => {
	const db = new Surreal();

	try {
		await db.connect(config.url);
		await db.use({ namespace: config.namespace, database: config.database });
		await db.signin({ username: 'root', password: 'root' });

		// const token = AuthCookie.get();
		// if (token) {
		//   try {
		//     await authenticateConnection(db, token);
		//   } catch {
		//     console.debug("No auth cookie");
		//     AuthCookie.remove();
		//   }
		// }

		return db;
	} catch (err) {
		console.error(
			'Failed to connect to SurrealDB:',
			err instanceof Error ? err.message : String(err)
		);
		await db.close();
		throw err;
	}
};

export const getDb = async (config: DbConfig = DEFAULT_CONFIG): Promise<Surreal> => {
	if (!connectionPromise) {
		connectionPromise = connectToDatabase(config);
	}

	return connectionPromise;
};

async function authenticateConnection(db: Surreal, token: string): Promise<void> {
	try {
		await db.authenticate(token);
		const userInfo = await db.info<User>();
		if (!userInfo) {
			throw new Error('Failed to obtain user info');
		}
		isAuthenticated = true;
		currentUser = userInfo;
	} catch (error) {
		console.error('Failed to authenticate connection:', error);
		isAuthenticated = false;
		currentUser = null;
		throw error;
	}
}

export async function invalidateSession(db: Surreal): Promise<void> {
	try {
		await db.invalidate();
		isAuthenticated = false;
		currentUser = null;
		AuthCookie.remove();
	} catch (error) {
		console.error('Failed to invalidate session:', error);
		throw error;
	}
}

export async function validateSession(db: Surreal, token: string): Promise<SessionInfo | null> {
	try {
		await db.authenticate(token);
		const [result] = await db.query('$session');
		const session = sessionInfoSchema.parse(result);
		return session;
	} catch (err) {
		console.error('Session validation failed:', err);
		await invalidateSession(db);
		return null;
	}
}

// TODO move
export async function refreshAuth() {
	try {
		const db = await getDb();
		const info = await db.info();
		const user = (await userSchema.parseAsync(jsonify(info))) as User;

		currentUser = user;
		isAuthenticated = true;
		return user;
	} catch (error) {
		console.error('Failed refreshing auth', error);
		currentUser = null;
		isAuthenticated = false;
		return null;
	}
}

// export const refreshAuth = async () => {
// 	const user = await fetchUser().catch(() => undefined);

// 	const result = {
// 		user,
// 		authenticated: !!user,
// 	};

// 	state.set({
// 		initialized: true,
// 		...result,
// 	});

// 	return result;
// };

// import { Surreal } from 'surrealdb';
// import type { SessionInfo, User } from '$lib/types';
// import { sessionInfoSchema } from './zod';

// class SurrealClient {
//   private static instance: SurrealClient;
//   private _db: Surreal;
//   public get db(): Surreal {
//     return this._db;
//   }

//   private _isInitialized = $state(false);
//   public get isInitialized() {
//     return this._isInitialized;
//   }
//   private _isAuthenticated = $state(false);
//   public get isAuthenticated() {
//     return this._isAuthenticated;
//   }
//   private _user: User | null = $state(null);
//   public get user(): User | null {
//     return this._user;
//   }

//   private constructor() {
//     this._db = new Surreal();
//   }

//   public static getInstance(): SurrealClient {
//     if (!SurrealClient.instance) {
//       SurrealClient.instance = new SurrealClient();
//     }
//     return SurrealClient.instance;
//   }

//   async initialize(dbUrl: string, namespace: string, database: string) {
//     if (this._isInitialized) return;

//     try {
//       await this._db.connect(dbUrl);
//       await this._db.use({ namespace, database });
//       this._isInitialized = true;
//     } catch (error) {
//       console.error('Failed to initialize database connection:', error);
//       throw error;
//     }
//   }

//   async authenticate(token: string) {
//     try {
//       await this._db.authenticate(token);
//       const userInfo = await this._db.info<User>();
//       // TODO technically only an error for record users, root users will have a valid session without an $auth record
//       if (!userInfo) throw new Error("Failed to obtain user info")
//       this._isAuthenticated = true
//       this._user = userInfo
//     } catch (error) {
//       console.error('Failed to authenticate connection:', error);
//       this._isAuthenticated = false
//       this._user = null
//       throw error;
//     }
//   }

//   async invalidate() {
//     try {
//       await this._db.invalidate();
//       this._isAuthenticated = false
//       this._user = null
//     } catch (error) {
//       console.error('Failed to invalidate session:', error);
//       throw error;
//     }
//   }

//   async validateSession(token: string): Promise<SessionInfo | null> {
//     try {
//       await this._db.authenticate(token);
//       // const result = await surreal.info<User>();
//       // const user = userSchema.parse(result);
//       // return user;
//       const [result] = await this._db.query("$session");
//       const session = sessionInfoSchema.parse(result);
//       return session;
//     } catch (err) {
//       console.error('Session validation failed:', err);
//       this.invalidate()
//       return null;
//     }
//   }
// }

// export const surreal = SurrealClient.getInstance();
