import z from 'zod';
import { RecordId, StringRecordId } from 'surrealdb';

type TableRecordId<T extends string> = RecordId<T> | StringRecordId | `${T}:${string}`;

export function recordId<Table extends string = string>(table?: Table) {
	const tableRegex = table ? table : '[A-Za-z_][A-Za-z0-9_]*';
	const idRegex = '[^:]+';
	const fullRegex = new RegExp(`^${tableRegex}:${idRegex}$`);

	return z
		.union([
			z
				.custom<RecordId<string>>((val): val is RecordId<string> => val instanceof RecordId)
				.refine((val): val is RecordId<Table> => !table || val.tb === table, {
					message: table ? `RecordId must be of type '${table}'` : undefined
				}),
			z
				.custom<StringRecordId>((val): val is StringRecordId => val instanceof StringRecordId)
				.refine((val) => !table || val.rid.startsWith(`${table}:`), {
					message: table ? `StringRecordId must start with '${table}:'` : undefined
				}),
			z.string().regex(fullRegex, {
				message: table
					? `Invalid record ID format. Must be '${table}:id'`
					: "Invalid record ID format. Must be 'table:id'"
			})
		])
		.transform((val): TableRecordId<Table> => {
			if (typeof val === 'string') {
				return new StringRecordId(val) as TableRecordId<Table>;
			}
			return val as TableRecordId<Table>;
		});
}

export const role = z.enum(['admin', 'reviewer', 'tutor', 'student']);
export const roles = z.union([z.array(role), z.set(role)]).transform((val) => {
	return val instanceof Set ? val : new Set(val);
});

export type Role = z.infer<typeof role>;
export type Roles = z.infer<typeof roles>;

// The select schema for table user
// const userRecordSchema = z.object({
// 	id: recordId('user'),
// 	name: z.string(),
// 	email: z.string().email(),
// 	enabled: z.boolean(),
// 	password_hash: z.string().optional(),
// 	roles: roles,
// 	activation_token: z.string().optional(),
// 	activation_expires: z.date().optional()
// });

// export const userSchema = userRecordSchema.transform((user) => ({
// 	...user,
// 	id: user.id.toString(),
// 	roles: Array.from(user.roles)
// }));

// export type User = z.infer<typeof userSchema>;

export const userSignUpSchema = z.object({
	name: z.string().min(3),
	email: z.string().email(),
	// enabled: z.boolean().optional(),
	password: z.string()
	// roles: roles.optional()
});

export type SignUpCredentials = z.infer<typeof userSignUpSchema>;

export const userSignInSchema = z.object({
	email: z.string().email(),
	password: z.string()
});

export type SignInCredentials = z.infer<typeof userSignInSchema>;

export const createUserSchema = z.object({
	name: z.string().min(3),
	email: z.string().email(),
	roles: z.array(role)
});

export const activateAccountSchema = z.object({
	token: z.string(),
	password: z.string().min(8)
});

export const deleteUserSchema = z.object({ userId: z.string() });

export type CreateUser = z.infer<typeof createUserSchema>;
export type DeleteUser = z.infer<typeof deleteUserSchema>;
export type ActivateAccount = z.infer<typeof activateAccountSchema>;

export const tokenInfoSchema = z.object({
	AC: z.string().optional(),
	DB: z.string().optional(),
	ID: z.string(),
	NS: z.string().optional(),
	exp: z.number(),
	iat: z.number(),
	iss: z.string(),
	jti: z.string(),
	nbf: z.number()
});

export type TokenInfo = z.infer<typeof tokenInfoSchema>;

export const sessionInfoSchema = z.object({
	ac: z.string().optional(),
	db: z.string(),
	ns: z.string(),
	exp: z.number().optional(),
	id: z.string(),
	ip: z.string(),
	or: z.string().optional(),
	rd: recordId('user').optional(),
	tk: tokenInfoSchema
});

export type SessionInfo = z.infer<typeof sessionInfoSchema>;
